import Vue from 'vue';
import { throttle } from './helpers';
import { Context } from '@idalite-frontend/nuxt-types';
import {
    createNuxtDevice,
    NuxtDevice,
    NuxtDeviceOptions,
    NuxtDeviceType,
    NuxtDeviceTypeEnum,
} from '~/types';

interface NuxtProcess extends NodeJS.Process {
    server: boolean;
    client: boolean;
}

declare let process: NuxtProcess;

const DEFAULT_OPTIONS: NuxtDeviceOptions = {
    mobile: 720,
    tablet: 1200,
    laptop: 1440,
    desktop: 1920,
};

function isMobile(str?: string): boolean {
    if (!str) {
        return false;
    }

    return /Mobile|Android|iP(hone|od)|IEMobile|BlackBerry|Kindle|Silk-Accelerated|(hpw|web)OS|Opera M(obi|ini)/.test(str); // eslint-disable-line
}

function isTablet(str?: string): boolean {
    if (!str) {
        return false;
    }

    return /(tablet|ipad|playbook|silk)|(android(?!.*mobi))/i.test(str); // eslint-disable-line
}

function getDeviceFromUA(str?: string): string {
    if (isMobile(str)) {
        return NuxtDeviceTypeEnum.Mobile;
    } else if (isTablet(str)) {
        return NuxtDeviceTypeEnum.Tablet;
    } else {
        return NuxtDeviceTypeEnum.Desktop;
    }
}

function getDeviceFromBreakpoints(container: HTMLElement, breakpoints: NuxtDeviceOptions): NuxtDeviceType {
    const { width } = container.getBoundingClientRect();

    const keys: Array<string> = Object.keys(breakpoints);
    const values: Array<number> = Object.values(breakpoints);

    const reversedBreakpoints: Record<number, string> = keys.reduce((acc, key) => ({
        ...acc,
        [breakpoints[key]]: key,
    }), {});

    const minBreakpoint: number = values
        .reduce((last, current) => width <= current && current < last ? current : last, values[values.length - 1]);

    return reversedBreakpoints[minBreakpoint];
}

function setNuxtAppBodyOpacity(ctx: Context): void {
    if (!ctx.app.head) {
        ctx.app.head = {};
    }

    if (!ctx.app.head.bodyAttrs) {
        ctx.app.head.bodyAttrs = {};
    }

    const bodyStyle: string = ctx.app.head.bodyAttrs.style || '';
    ctx.app.head.bodyAttrs.style = 'opacity: 0;' + bodyStyle.replace(/opacity:\s?\.?\d*;?/g, '');
}

function resetNuxtAppBodyOpacity(ctx: Context): void {
    if (!ctx.app.head) {
        ctx.app.head = {};
    }

    if (!ctx.app.head.bodyAttrs) {
        ctx.app.head.bodyAttrs = {};
    }

    const bodyStyle: string = ctx.app.head.bodyAttrs.style || '';
    ctx.app.head.bodyAttrs.style = bodyStyle.replace(/opacity:\s?\.?\d*;?/g, '');
}

function setDevice(instance: NuxtDevice, breakpoints: NuxtDeviceOptions): void {
    const result: NuxtDeviceType = getDeviceFromBreakpoints(document.body, breakpoints);

    if (instance.state === result) {
        return;
    }

    instance.state = result;
}

function initClientSide(instance: NuxtDevice, breakpoints: NuxtDeviceOptions) {
    document.body.style.opacity = '';

    setDevice(instance, breakpoints);
    window.addEventListener('resize', throttle(100, () => setDevice(instance, breakpoints)));
}

function redefineAppMounted(ctx: Context, instance: NuxtDevice, breakpoints: NuxtDeviceOptions): void {
    const _mounted = ctx.app.mounted;

    ctx.app.mounted = () => {
        setTimeout(() => {
            initClientSide(instance, breakpoints);

            if (_mounted) {
                _mounted.bind(ctx.app);
            }
        }, 300);
    };
}

const createNuxtDeviceInstance: createNuxtDevice = (ctx: Context, breakpoints: NuxtDeviceOptions = DEFAULT_OPTIONS): NuxtDevice => {
    let defaultState: NuxtDeviceType = NuxtDeviceTypeEnum.Mobile;

    if (process.server) {
        defaultState = getDeviceFromUA(ctx.req.headers['user-agent']);
    } else if (process.client) {
        defaultState = getDeviceFromUA(window.navigator.userAgent);
    }

    const instance: NuxtDevice = Vue.observable({
        state: defaultState,
        is: function(...args: Array<NuxtDeviceType>): boolean {
            return args.includes(this.state);
        },
        isMobile: function() {
            return this.state === NuxtDeviceTypeEnum.Mobile;
        },
        isTablet: function() {
            return this.state === NuxtDeviceTypeEnum.Tablet;
        },
        isLaptop: function() {
            return this.state === NuxtDeviceTypeEnum.Laptop;
        },
        isDesktop: function() {
            return this.state === NuxtDeviceTypeEnum.Desktop;
        },
    });

    if (process.server) {
        setNuxtAppBodyOpacity(ctx);
    } else if (process.client) {
        resetNuxtAppBodyOpacity(ctx);
        redefineAppMounted(ctx, instance, breakpoints);
    }

    return instance;
};

export default createNuxtDeviceInstance;
